"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var get_stream_1 = require("./get-stream");
var _ = require("lodash");
var immutable_1 = require("immutable");
var targetLanguages = require("./Language/All");
var Support_1 = require("./Support");
var CompressedJSON_1 = require("./CompressedJSON");
var CombineClasses_1 = require("./CombineClasses");
var JSONSchemaInput_1 = require("./JSONSchemaInput");
var Inference_1 = require("./Inference");
var InferMaps_1 = require("./InferMaps");
var TypeBuilder_1 = require("./TypeBuilder");
var TypeGraph_1 = require("./TypeGraph");
var GraphQL_1 = require("./GraphQL");
var GatherNames_1 = require("./GatherNames");
var TypeNames_1 = require("./TypeNames");
var InferEnums_1 = require("./InferEnums");
// Re-export essential types and functions
var TargetLanguage_1 = require("./TargetLanguage");
exports.TargetLanguage = TargetLanguage_1.TargetLanguage;
var All_1 = require("./Language/All");
exports.languages = All_1.all;
exports.languageNamed = All_1.languageNamed;
var stringToStream = require("string-to-stream");
function getTargetLanguage(name) {
    var language = targetLanguages.languageNamed(name);
    if (language !== undefined) {
        return language;
    }
    throw new Error("'" + name + "' is not yet supported as an output language.");
}
exports.getTargetLanguage = getTargetLanguage;
function isJSONSource(source) {
    return "samples" in source;
}
exports.isJSONSource = isJSONSource;
function isSchemaSource(source) {
    return !("query" in source) && !("samples" in source);
}
exports.isSchemaSource = isSchemaSource;
function isGraphQLSource(source) {
    return "query" in source;
}
exports.isGraphQLSource = isGraphQLSource;
var defaultOptions = {
    lang: "ts",
    sources: [],
    handlebarsTemplate: undefined,
    inferMaps: true,
    inferEnums: true,
    alphabetizeProperties: false,
    combineClasses: true,
    noRender: false,
    leadingComments: undefined,
    rendererOptions: {},
    indentation: undefined,
    outputFilename: "stdout"
};
function toReadable(source) {
    return _.isString(source) ? stringToStream(source) : source;
}
function toString(source) {
    return __awaiter(this, void 0, void 0, function () {
        var _a;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    if (!_.isString(source)) return [3 /*break*/, 1];
                    _a = source;
                    return [3 /*break*/, 3];
                case 1: return [4 /*yield*/, get_stream_1.getStream(source)];
                case 2:
                    _a = _b.sent();
                    _b.label = 3;
                case 3: return [2 /*return*/, _a];
            }
        });
    });
}
var Run = /** @class */ (function () {
    function Run(options) {
        var _this = this;
        this.makeGraph = function () {
            var targetLanguage = getTargetLanguage(_this._options.lang);
            var stringTypeMapping = targetLanguage.stringTypeMapping;
            var typeBuilder = new TypeBuilder_1.TypeGraphBuilder(stringTypeMapping, _this._options.alphabetizeProperties);
            // JSON Schema
            immutable_1.Map(_this._allInputs.schemas).forEach(function (schema, name) {
                typeBuilder.addTopLevel(name, JSONSchemaInput_1.schemaToType(typeBuilder, name, schema));
            });
            // GraphQL
            var numInputs = Object.keys(_this._allInputs.graphQLs).length;
            immutable_1.Map(_this._allInputs.graphQLs).forEach(function (_a, name) {
                var schema = _a.schema, query = _a.query;
                var newTopLevels = GraphQL_1.makeGraphQLQueryTypes(name, typeBuilder, schema, query);
                newTopLevels.forEach(function (t, actualName) {
                    typeBuilder.addTopLevel(numInputs === 1 ? name : actualName, t);
                });
            });
            // JSON
            var doInferEnums = _this._options.inferEnums;
            if (Object.keys(_this._allInputs.samples).length > 0) {
                var inference_1 = new Inference_1.TypeInference(typeBuilder, doInferEnums);
                immutable_1.Map(_this._allInputs.samples).forEach(function (cjson, name) {
                    typeBuilder.addTopLevel(name, inference_1.inferType(_this._compressedJSON, TypeNames_1.makeTypeNames(name, false), cjson));
                });
            }
            var originalGraph = typeBuilder.finish();
            var graph = originalGraph;
            if (_this._options.combineClasses) {
                graph = CombineClasses_1.combineClasses(graph, stringTypeMapping, _this._options.alphabetizeProperties);
            }
            if (doInferEnums) {
                graph = InferEnums_1.inferEnums(graph, stringTypeMapping);
            }
            if (_this._options.inferMaps) {
                graph = InferMaps_1.inferMaps(graph, stringTypeMapping);
            }
            graph = TypeGraph_1.noneToAny(graph, stringTypeMapping);
            if (!targetLanguage.supportsOptionalClassProperties) {
                graph = TypeGraph_1.optionalToNullable(graph, stringTypeMapping);
            }
            // JSON Schema input can leave unreachable classes in the graph when it
            // unifies, which can trip is up, so we remove them here.  Also, sometimes
            // we combine classes in ways that will the order come out differently
            // compared to what it would be from the equivalent schema, so we always
            // just garbage collect to get a defined order and be done with it.
            // FIXME: We don't actually have to do this if any of the above graph
            // rewrites did anything.  We could just check whether the current graph
            // is different from the one we started out with.
            graph = graph.garbageCollect();
            GatherNames_1.gatherNames(graph);
            return graph;
        };
        this.run = function () { return __awaiter(_this, void 0, void 0, function () {
            var targetLanguage, _i, _a, source, name, schema, query, _b, _c, _d, name, samples, _e, samples_1, sample, input, name, schema, input, _f, _g, graph;
            return __generator(this, function (_h) {
                switch (_h.label) {
                    case 0:
                        targetLanguage = getTargetLanguage(this._options.lang);
                        _i = 0, _a = this._options.sources;
                        _h.label = 1;
                    case 1:
                        if (!(_i < _a.length)) return [3 /*break*/, 12];
                        source = _a[_i];
                        if (!isGraphQLSource(source)) return [3 /*break*/, 3];
                        name = source.name, schema = source.schema, query = source.query;
                        _b = this._allInputs.graphQLs;
                        _c = name;
                        _d = { schema: schema };
                        return [4 /*yield*/, toString(query)];
                    case 2:
                        _b[_c] = (_d.query = _h.sent(), _d);
                        return [3 /*break*/, 11];
                    case 3:
                        if (!isJSONSource(source)) return [3 /*break*/, 8];
                        name = source.name, samples = source.samples;
                        _e = 0, samples_1 = samples;
                        _h.label = 4;
                    case 4:
                        if (!(_e < samples_1.length)) return [3 /*break*/, 7];
                        sample = samples_1[_e];
                        return [4 /*yield*/, this._compressedJSON.readFromStream(toReadable(sample))];
                    case 5:
                        input = _h.sent();
                        if (!_.has(this._allInputs.samples, name)) {
                            this._allInputs.samples[name] = [];
                        }
                        this._allInputs.samples[name].push(input);
                        _h.label = 6;
                    case 6:
                        _e++;
                        return [3 /*break*/, 4];
                    case 7: return [3 /*break*/, 11];
                    case 8:
                        if (!isSchemaSource(source)) return [3 /*break*/, 10];
                        name = source.name, schema = source.schema;
                        _g = (_f = JSON).parse;
                        return [4 /*yield*/, toString(schema)];
                    case 9:
                        input = _g.apply(_f, [_h.sent()]);
                        if (_.has(this._allInputs.schemas, name)) {
                            throw new Error("More than one schema given for " + name);
                        }
                        this._allInputs.schemas[name] = input;
                        return [3 /*break*/, 11];
                    case 10:
                        Support_1.assertNever(source);
                        _h.label = 11;
                    case 11:
                        _i++;
                        return [3 /*break*/, 1];
                    case 12:
                        graph = this.makeGraph();
                        if (this._options.noRender) {
                            return [2 /*return*/, immutable_1.OrderedMap([[this._options.outputFilename, { lines: ["Done.", ""], annotations: immutable_1.List() }]])];
                        }
                        if (this._options.handlebarsTemplate !== undefined) {
                            return [2 /*return*/, immutable_1.OrderedMap([
                                    [
                                        this._options.outputFilename,
                                        targetLanguage.processHandlebarsTemplate(graph, this._options.rendererOptions, this._options.handlebarsTemplate)
                                    ]
                                ])];
                        }
                        else {
                            return [2 /*return*/, targetLanguage.renderGraphAndSerialize(graph, this._options.outputFilename, this._options.alphabetizeProperties, this._options.leadingComments, this._options.rendererOptions, this._options.indentation)];
                        }
                        return [2 /*return*/];
                }
            });
        }); };
        this._options = _.assign(_.clone(defaultOptions), options);
        this._allInputs = { samples: {}, schemas: {}, graphQLs: {} };
        var mapping = getTargetLanguage(this._options.lang).stringTypeMapping;
        var makeDate = mapping.date !== "string";
        var makeTime = mapping.time !== "string";
        var makeDateTime = mapping.dateTime !== "string";
        this._compressedJSON = new CompressedJSON_1.CompressedJSON(makeDate, makeTime, makeDateTime);
    }
    return Run;
}());
exports.Run = Run;
function quicktypeMultiFile(options) {
    return new Run(options).run();
}
exports.quicktypeMultiFile = quicktypeMultiFile;
function offsetLocation(loc, lineOffset) {
    return { line: loc.line + lineOffset, column: loc.column };
}
function offsetSpan(span, lineOffset) {
    return { start: offsetLocation(span.start, lineOffset), end: offsetLocation(span.end, lineOffset) };
}
function quicktype(options) {
    return __awaiter(this, void 0, void 0, function () {
        var result, first, lines, annotations;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, quicktypeMultiFile(options)];
                case 1:
                    result = _a.sent();
                    if (result.size <= 1) {
                        first = result.first();
                        if (first === undefined) {
                            return [2 /*return*/, { lines: [], annotations: immutable_1.List() }];
                        }
                        return [2 /*return*/, first];
                    }
                    lines = [];
                    annotations = [];
                    result.forEach(function (srr, filename) {
                        var offset = lines.length + 2;
                        lines = lines.concat(["// " + filename, ""], srr.lines);
                        annotations = annotations.concat(srr.annotations.map(function (ann) { return ({ annotation: ann.annotation, span: offsetSpan(ann.span, offset) }); }).toArray());
                    });
                    return [2 /*return*/, { lines: lines, annotations: immutable_1.List(annotations) }];
            }
        });
    });
}
exports.quicktype = quicktype;
